@@ -31,13 +31,15 @@ import collections
 import dataclasses
 import re
 from itertools import chain
-from typing import (cast, Dict, FrozenSet, List, Mapping, Optional,
-        Sequence, Set, Union)
+from typing import (cast, Dict, FrozenSet, Iterable, List, Mapping, Optional,
+                    Sequence, Set, Union)
 
-from google.api import annotations_pb2  # type: ignore
+from google.api import annotations_pb2      # type: ignore
 from google.api import client_pb2
 from google.api import field_behavior_pb2
-from google.protobuf import descriptor_pb2
+from google.api import resource_pb2
+from google.api_core import exceptions      # type: ignore
+from google.protobuf import descriptor_pb2  # type: ignore
 
 from gapic import utils
 from gapic.schema import metadata
@@ -69,6 +71,11 @@ class Field:
         """Return True if the field is a primitive, False otherwise."""
         return isinstance(self.type, PrimitiveType)
 
+    @property
+    def map(self) -> bool:
+        """Return True if this field is a map, False otherwise."""
+        return bool(self.repeated and self.message and self.message.map)
+
     @utils.cached_property
     def mock_value(self) -> str:
         """Return a repr of a valid, usually truthy mock value."""
@@ -103,9 +110,16 @@ class Field:
             sub = next(iter(self.type.fields.values()))
             answer = f'{self.type.ident}({sub.name}={sub.mock_value})'
 
-        # If this is a repeated field, then the mock answer should
-        # be a list.
-        if self.repeated:
+        if self.map:
+            # Maps are a special case beacuse they're represented internally as
+            # a list of a generated type with two fields: 'key' and 'value'.
+            answer = '{{{}: {}}}'.format(
+                self.type.fields["key"].mock_value,
+                self.type.fields["value"].mock_value,
+            )
+        elif self.repeated:
+            # If this is a repeated field, then the mock answer should
+            # be a list.
             answer = f'[{answer}]'
 
         # Done; return the mock value.
@@ -126,7 +140,8 @@ class Field:
             bool: Whether this field is repeated.
         """
         return self.label == \
-            descriptor_pb2.FieldDescriptorProto.Label.Value(b'LABEL_REPEATED')
+            descriptor_pb2.FieldDescriptorProto.Label.Value(
+                'LABEL_REPEATED')   # type: ignore
 
     @property
     def required(self) -> bool:
@@ -169,8 +184,8 @@ class Field:
             return PrimitiveType.build(bytes)
 
         # This should never happen.
-        raise TypeError('Unrecognized protobuf type. This code should '
-                        'not be reachable; please file a bug.')
+        raise TypeError(f'Unrecognized protobuf type: {self.field_pb.type}. '
+                        'This code should not be reachable; please file a bug.')
 
     def with_context(self, *, collisions: FrozenSet[str]) -> 'Field':
         """Return a derivative of this field with the provided context.
@@ -179,7 +194,8 @@ class Field:
         ``Field`` object aliases module names to avoid naming collisions
         in the file being written.
         """
-        return dataclasses.replace(self,
+        return dataclasses.replace(
+            self,
             message=self.message.with_context(
                 collisions=collisions,
                 skip_fields=True,
@@ -204,22 +220,69 @@ class MessageType:
     def __getattr__(self, name):
         return getattr(self.message_pb, name)
 
+    def __hash__(self):
+        # Identity is sufficiently unambiguous.
+        return hash(self.ident)
+
     @utils.cached_property
     def field_types(self) -> Sequence[Union['MessageType', 'EnumType']]:
+        answer = tuple(
+            field.type
+            for field in self.fields.values()
+            if field.message or field.enum
+        )
+
+        return answer
+
+    @utils.cached_property
+    def recursive_field_types(self) -> Sequence[
+        Union['MessageType', 'EnumType']
+    ]:
         """Return all composite fields used in this proto's messages."""
-        answer = []
-        for field in self.fields.values():
-            if field.message or field.enum:
-                answer.append(field.type)
-        return tuple(answer)
+        types: Set[Union['MessageType', 'EnumType']] = set()
+
+        stack = [iter(self.fields.values())]
+        while stack:
+            fields_iter = stack.pop()
+            for field in fields_iter:
+                if field.message and field.type not in types:
+                    stack.append(iter(field.message.fields.values()))
+                if not field.is_primitive:
+                    types.add(field.type)
+
+        return tuple(types)
+
+    @property
+    def map(self) -> bool:
+        """Return True if the given message is a map, False otherwise."""
+        return self.message_pb.options.map_entry
 
     @property
     def ident(self) -> metadata.Address:
         """Return the identifier data to be used in templates."""
         return self.meta.address
 
+    @property
+    def resource_path(self) -> Optional[str]:
+        """If this message describes a resource, return the path to the resource.
+        If there are multiple paths, returns the first one."""
+        return next(
+            iter(self.options.Extensions[resource_pb2.resource].pattern),
+            None
+        )
+
+    @property
+    def resource_type(self) -> Optional[str]:
+        resource = self.options.Extensions[resource_pb2.resource]
+        return resource.type[resource.type.find('/') + 1:] if resource else None
+
+    @property
+    def resource_path_args(self) -> Sequence[str]:
+        path_arg_re = re.compile(r'\{([a-zA-Z0-9_-]+)\}')
+        return path_arg_re.findall(self.resource_path or '')
+
     def get_field(self, *field_path: str,
-            collisions: FrozenSet[str] = frozenset()) -> Field:
+                  collisions: FrozenSet[str] = frozenset()) -> Field:
         """Return a field arbitrarily deep in this message's structure.
 
         This method recursively traverses the message tree to return the
@@ -277,8 +340,8 @@ class MessageType:
         return cursor.message.get_field(*field_path[1:], collisions=collisions)
 
     def with_context(self, *,
-            collisions: FrozenSet[str],
-            skip_fields: bool = False,
+                     collisions: FrozenSet[str],
+                     skip_fields: bool = False,
                      ) -> 'MessageType':
         """Return a derivative of this message with the provided context.
 
@@ -290,19 +353,21 @@ class MessageType:
         underlying fields. This provides for an "exit" in the case of circular
         references.
         """
-        return dataclasses.replace(self,
-            fields=collections.OrderedDict([
+        return dataclasses.replace(
+            self,
+            fields=collections.OrderedDict(
                 (k, v.with_context(collisions=collisions))
                 for k, v in self.fields.items()
-            ]) if not skip_fields else self.fields,
-            nested_enums=collections.OrderedDict([
+            ) if not skip_fields else self.fields,
+            nested_enums=collections.OrderedDict(
                 (k, v.with_context(collisions=collisions))
                 for k, v in self.nested_enums.items()
-            ]),
-            nested_messages=collections.OrderedDict([(k, v.with_context(
-                collisions=collisions,
-                skip_fields=skip_fields,
-            )) for k, v in self.nested_messages.items()]),
+            ),
+            nested_messages=collections.OrderedDict(
+                (k, v.with_context(
+                    collisions=collisions,
+                    skip_fields=skip_fields,))
+                for k, v in self.nested_messages.items()),
             meta=self.meta.with_context(collisions=collisions),
         )
 
@@ -328,9 +393,20 @@ class EnumType:
         default_factory=metadata.Metadata,
     )
 
+    def __hash__(self):
+        # Identity is sufficiently unambiguous.
+        return hash(self.ident)
+
     def __getattr__(self, name):
         return getattr(self.enum_pb, name)
 
+    @property
+    def resource_path(self) -> Optional[str]:
+        # This is a minor duck-typing workaround for the resource_messages
+        # property in the Service class: we need to check fields recursively
+        # to see if they're resources, and recursive_field_types includes enums
+        return None
+
     @property
     def ident(self) -> metadata.Address:
         """Return the identifier data to be used in templates."""
@@ -343,9 +419,10 @@ class EnumType:
         ``EnumType`` object aliases module names to avoid naming collisions in
         the file being written.
         """
-        return dataclasses.replace(self,
+        return dataclasses.replace(
+            self,
             meta=self.meta.with_context(collisions=collisions),
-                                   )
+        )
 
 
 @dataclasses.dataclass(frozen=True)
@@ -373,6 +450,10 @@ class PythonType:
     def name(self) -> str:
         return self.ident.name
 
+    @property
+    def field_types(self) -> Sequence[Union['MessageType', 'EnumType']]:
+        return tuple()
+
 
 @dataclasses.dataclass(frozen=True)
 class PrimitiveType(PythonType):
@@ -411,6 +492,33 @@ class OperationInfo:
     response_type: MessageType
     metadata_type: MessageType
 
+    def with_context(self, *, collisions: FrozenSet[str]) -> 'OperationInfo':
+        """Return a derivative of this OperationInfo with the provided context.
+
+          This method is used to address naming collisions. The returned
+          ``OperationInfo`` object aliases module names to avoid naming collisions
+          in the file being written.
+          """
+        return dataclasses.replace(
+            self,
+            response_type=self.response_type.with_context(
+                collisions=collisions
+            ),
+            metadata_type=self.metadata_type.with_context(
+                collisions=collisions
+            ),
+        )
+
+
+@dataclasses.dataclass(frozen=True)
+class RetryInfo:
+    """Representation of the method's retry behavior."""
+    max_attempts: int
+    initial_backoff: float
+    max_backoff: float
+    backoff_multiplier: float
+    retryable_exceptions: FrozenSet[exceptions.GoogleAPICallError]
+
 
 @dataclasses.dataclass(frozen=True)
 class Method:
@@ -419,6 +527,8 @@ class Method:
     input: MessageType
     output: MessageType
     lro: Optional[OperationInfo] = dataclasses.field(default=None)
+    retry: Optional[RetryInfo] = dataclasses.field(default=None)
+    timeout: Optional[float] = None
     meta: metadata.Metadata = dataclasses.field(
         default_factory=metadata.Metadata,
     )
@@ -466,8 +576,7 @@ class Method:
             return PythonType(meta=metadata.Metadata(
                 address=metadata.Address(
                     name=f'{self.name}Pager',
-                    package=self.ident.api_naming.module_namespace + (
-                        self.ident.api_naming.versioned_module_name,
+                    package=self.ident.api_naming.module_namespace + (self.ident.api_naming.versioned_module_name,) + self.ident.subpackage + (
                         'services',
                         utils.to_snake_case(self.ident.parent[-1]),
                     ),
@@ -495,28 +604,40 @@ class Method:
     @utils.cached_property
     def flattened_fields(self) -> Mapping[str, Field]:
         """Return the signature defined for this method."""
-        answer: Dict[str, Field] = collections.OrderedDict()
+        cross_pkg_request = self.input.ident.package != self.ident.package
+
+        def filter_fields(sig):
+            for f in sig.split(','):
+                if not f:
+                    # Special case for an empty signature
+                    continue
+                name = f.strip()
+                field = self.input.get_field(*name.split('.'))
+                if cross_pkg_request and not field.is_primitive:
+                    # This is not a proto-plus wrapped message type,
+                    # and setting a non-primitive field directly is verboten.
+                    continue
+
+                yield name, field
+
         signatures = self.options.Extensions[client_pb2.method_signature]
+        answer: Dict[str, Field] = collections.OrderedDict(
+            name_and_field
+            for sig in signatures
+            for name_and_field in filter_fields(sig)
+        )
 
-        # Iterate over each signature and add the appropriate fields.
-        for sig in signatures:
-            # Get all of the individual fields.
-            fields = collections.OrderedDict([
-                (f, self.input.get_field(*f.split('.')))
-                for f in sig.split(',')
-            ])
-
-            # Sanity check: If any fields contain a message, we ignore the
-            # entire signature.
-            if any([i.message for i in fields.values()]):
-                continue
-
-            # Add the fields to the answer.
-            answer.update(fields)
-
-        # Done; return the flattened fields
         return answer
 
+    @utils.cached_property
+    def legacy_flattened_fields(self) -> Mapping[str, Field]:
+        """Return the legacy flattening interface: top level fields only,
+        required fields first"""
+        required, optional = utils.partition(lambda f: f.required,
+                                             self.input.fields.values())
+        return collections.OrderedDict((f.name, f)
+                                       for f in chain(required, optional))
+
     @property
     def grpc_stub_type(self) -> str:
         """Return the type of gRPC stub to use."""
@@ -562,20 +683,29 @@ class Method:
 
     @utils.cached_property
     def ref_types(self) -> Sequence[Union[MessageType, EnumType]]:
+        return self._ref_types(True)
+
+    @utils.cached_property
+    def flat_ref_types(self) -> Sequence[Union[MessageType, EnumType]]:
+        return self._ref_types(False)
+
+    def _ref_types(self, recursive: bool) -> Sequence[Union[MessageType, EnumType]]:
         """Return types referenced by this method."""
         # Begin with the input (request) and output (response) messages.
-        answer = [self.input]
+        answer: List[Union[MessageType, EnumType]] = [self.input]
+        types: Iterable[Union[MessageType, EnumType]] = (
+            self.input.recursive_field_types if recursive
+            else (
+                f.type
+                for f in self.flattened_fields.values()
+                if f.message or f.enum
+            )
+        )
+        answer.extend(types)
+
         if not self.void:
             answer.append(self.client_output)
-
-        # If this method has flattening that is honored, add its
-        # composite types.
-        #
-        # This entails adding the module for any field on the signature
-        # unless the field is a primitive.
-        for field in self.flattened_fields.values():
-            if field.message or field.enum:
-                answer.append(field.type)
+            answer.extend(self.client_output.field_types)
 
         # If this method has LRO, it is possible (albeit unlikely) that
         # the LRO messages reside in a different module.
@@ -583,6 +713,11 @@ class Method:
             answer.append(self.lro.response_type)
             answer.append(self.lro.metadata_type)
 
+        # If this message paginates its responses, it is possible
+        # that the individual result messages reside in a different module.
+        if self.paged_result_field:
+            answer.append(self.paged_result_field.message)
+
         # Done; return the answer.
         return tuple(answer)
 
@@ -598,11 +733,17 @@ class Method:
         ``Method`` object aliases module names to avoid naming collisions
         in the file being written.
         """
-        return dataclasses.replace(self,
+        maybe_lro = self.lro.with_context(
+            collisions=collisions
+        ) if self.lro else None
+
+        return dataclasses.replace(
+            self,
+            lro=maybe_lro,
             input=self.input.with_context(collisions=collisions),
             output=self.output.with_context(collisions=collisions),
             meta=self.meta.with_context(collisions=collisions),
-                                   )
+        )
 
 
 @dataclasses.dataclass(frozen=True)
@@ -617,6 +758,19 @@ class Service:
     def __getattr__(self, name):
         return getattr(self.service_pb, name)
 
+    @property
+    def client_name(self) -> str:
+        """Returns the name of the generated client class"""
+        return self.name + "Client"
+
+    @property
+    def transport_name(self):
+        return self.name + "Transport"
+
+    @property
+    def grpc_transport_name(self):
+        return self.name + "GrpcTransport"
+
     @property
     def has_lro(self) -> bool:
         """Return whether the service has a long-running method."""
@@ -641,9 +795,11 @@ class Service:
             Sequence[str]: A sequence of OAuth scopes.
         """
         # Return the OAuth scopes, split on comma.
-        return tuple([i.strip() for i in
-            self.options.Extensions[client_pb2.oauth_scopes].split(',')
-            if i])
+        return tuple(
+            i.strip()
+            for i in self.options.Extensions[client_pb2.oauth_scopes].split(',')
+            if i
+        )
 
     @property
     def module_name(self) -> str:
@@ -662,23 +818,53 @@ class Service:
         used for imports.
         """
         # Put together a set of the service and method names.
-        answer = {self.name}.union(
-            {utils.to_snake_case(i.name) for i in self.methods.values()}
+        answer = {self.name, self.client_name}
+        answer.update(
+            utils.to_snake_case(i.name) for i in self.methods.values()
         )
 
         # Identify any import module names where the same module name is used
         # from distinct packages.
-        modules: Dict[str, Set[str]] = {}
-        for t in chain(*[m.ref_types for m in self.methods.values()]):
-            modules.setdefault(t.ident.module, set())
-            modules[t.ident.module].add(t.ident.package)
-        for module_name, packages in modules.items():
-            if len(packages) > 1:
-                answer.add(module_name)
+        modules: Dict[str, Set[str]] = collections.defaultdict(set)
+        for m in self.methods.values():
+            for t in m.ref_types:
+                modules[t.ident.module].add(t.ident.package)
+
+        answer.update(
+            module_name
+            for module_name, packages in modules.items()
+            if len(packages) > 1
+        )
 
         # Done; return the answer.
         return frozenset(answer)
 
+    @utils.cached_property
+    def resource_messages(self) -> FrozenSet[MessageType]:
+        """Returns all the resource message types used in all
+        request fields in the service."""
+        def gen_resources(message):
+            if message.resource_path:
+                yield message
+
+            for type_ in message.recursive_field_types:
+                if type_.resource_path:
+                    yield type_
+
+        return frozenset(
+            resource_msg
+            for method in self.methods.values()
+            for resource_msg in gen_resources(method.input)
+        )
+
+    @utils.cached_property
+    def any_client_streaming(self) -> bool:
+        return any(m.client_streaming for m in self.methods.values())
+
+    @utils.cached_property
+    def any_server_streaming(self) -> bool:
+        return any(m.server_streaming for m in self.methods.values())
+
     def with_context(self, *, collisions: FrozenSet[str]) -> 'Service':
         """Return a derivative of this service with the provided context.
 
@@ -686,10 +872,15 @@ class Service:
         ``Service`` object aliases module names to avoid naming collisions
         in the file being written.
         """
-        return dataclasses.replace(self,
-            methods=collections.OrderedDict([
-                (k, v.with_context(collisions=collisions))
+        return dataclasses.replace(
+            self,
+            methods=collections.OrderedDict(
+                (k, v.with_context(
+                    # A methodd's flattened fields create additional names
+                    # that may conflict with module imports.
+                    collisions=collisions | frozenset(v.flattened_fields.keys()))
+                 )
                 for k, v in self.methods.items()
-            ]),
+            ),
             meta=self.meta.with_context(collisions=collisions),
         )
