@@ -22,6 +22,11 @@ from google.protobuf import descriptor_pb2
 from gapic.schema import metadata
 from gapic.schema import wrappers
 
+from test_utils.test_utils import (
+    make_field,
+    make_message,
+)
+
 
 def test_field_properties():
     field = make_field(name='my_field', number=1, type='TYPE_BOOL')
@@ -173,6 +178,25 @@ def test_mock_value_repeated():
     assert field.mock_value == "['foo_bar_value']"
 
 
+def test_mock_value_map():
+    entry_msg = make_message(
+        name='SquidEntry',
+        fields=(
+            make_field(name='key', type='TYPE_STRING'),
+            make_field(name='value', type='TYPE_STRING'),
+        ),
+        options=descriptor_pb2.MessageOptions(map_entry=True),
+    )
+    field = make_field(
+        name='squids',
+        type_name='mollusc.SquidEntry',
+        message=entry_msg,
+        label=3,
+        type='TYPE_MESSAGE',
+    )
+    assert field.mock_value == "{'key_value': 'value_value'}"
+
+
 def test_mock_value_enum():
     values = [
         descriptor_pb2.EnumValueDescriptorProto(name='UNSPECIFIED', number=0),
@@ -217,14 +241,3 @@ def test_mock_value_message():
         message=message,
     )
     assert field.mock_value == 'bogus.Message(foo=324)'
-
-
-def make_field(*, message=None, enum=None, **kwargs) -> wrappers.Field:
-    T = descriptor_pb2.FieldDescriptorProto.Type
-    kwargs.setdefault('name', 'my_field')
-    kwargs.setdefault('number', 1)
-    kwargs.setdefault('type', T.Value('TYPE_BOOL'))
-    if isinstance(kwargs['type'], str):
-        kwargs['type'] = T.Value(kwargs['type'])
-    field_pb = descriptor_pb2.FieldDescriptorProto(**kwargs)
-    return wrappers.Field(field_pb=field_pb, message=message, enum=enum)
